"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Use = exports.Options = exports.Delete = exports.Put = exports.Patch = exports.Post = exports.Get = exports.ChildController = exports.Controller = void 0;
const express_1 = require("express");
const helpers_1 = require("../lib/helpers");
const enums_1 = require("./enums");
const globals_1 = require("../lib/globals");
const Controller = (prefix = "/", options) => (constructor) => {
    // Get Parent Router from Metadata
    const ChildControllers = (options === null || options === void 0 ? void 0 : options.childs) instanceof Array
        ? [
            ...options.childs
                .map((child) => Reflect.getMetadata(enums_1.MetaDataKeys.CONTROLLER, child))
                .filter((v) => v),
        ]
        : [];
    // Create Target Router
    const TargetRouter = express_1.Router(options === null || options === void 0 ? void 0 : options.routerOptions);
    // Save Controller to Metadata
    Reflect.defineMetadata(enums_1.MetaDataKeys.CONTROLLER, {
        prefix,
        router: TargetRouter,
        constructor,
        options,
    }, constructor);
    // Add Middlewares
    const Middlewares = Reflect.getMetadata(enums_1.MetaDataKeys.MIDDLEWARES, constructor) || [];
    if (Middlewares.length)
        TargetRouter.use(...Middlewares);
    // Push Child Controllers
    ChildControllers.forEach((controller) => TargetRouter.use(controller.prefix, controller.router));
    // Add Sub Router To Main Or Parent Router Instance
    if (!(options === null || options === void 0 ? void 0 : options.isChild))
        globals_1.EpicGlobals.getRouter().use(prefix, TargetRouter);
    // Get Routes
    const Routes = Reflect.getMetadata(enums_1.MetaDataKeys.ROUTES, constructor.prototype) || [];
    // Loop over Each Route
    Routes.forEach((Route) => {
        // Push Route Handler Details To Globals
        globals_1.EpicGlobals.pushRouteDetails({
            name: Route.name,
            description: Route.description,
        });
        // Add Route to Application
        TargetRouter[Route.method](Route.path, 
        // Add Request Name
        (req, _, next) => {
            // @ts-ignore
            req.name = Route.name;
            next();
        }, 
        // Add Route Middlewares
        ...(Reflect.getMetadata(enums_1.MetaDataKeys.ROUTE_MIDDLEWARES, constructor.prototype, Route.name) || []), 
        // Check Permissions
        helpers_1.EpicHelpers.requestHandler((req, res, next) => {
            // @ts-ignore
            if (!(options === null || options === void 0 ? void 0 : options.isOpen) && req.permissions instanceof Array) {
                // @ts-ignore
                if (!req.permissions.includes(req.name)) {
                    res.status(401);
                    return next(new Error(`You are not permitted!`));
                }
            }
            next();
        }), helpers_1.EpicHelpers.requestHandler(Route.handler));
    });
};
exports.Controller = Controller;
const ChildController = (prefix = "/", options) => exports.Controller(prefix, Object.assign(Object.assign({}, options), { isChild: true }));
exports.ChildController = ChildController;
// Create Route Binder Function
const RouteBinder = (method) => (path = "/", description = "No description has been provided!") => (target, key, desc) => {
    if (typeof desc.value === "function")
        Reflect.defineMetadata(enums_1.MetaDataKeys.ROUTES, [
            ...(Reflect.getMetadata(enums_1.MetaDataKeys.ROUTES, target) || []),
            {
                path,
                method,
                name: key,
                description,
                handler: desc.value,
            },
        ], target);
};
// Available Request Method Decorators
exports.Get = RouteBinder(enums_1.RequestMethods.GET);
exports.Post = RouteBinder(enums_1.RequestMethods.POST);
exports.Patch = RouteBinder(enums_1.RequestMethods.PATCH);
exports.Put = RouteBinder(enums_1.RequestMethods.PUT);
exports.Delete = RouteBinder(enums_1.RequestMethods.DELETE);
exports.Options = RouteBinder(enums_1.RequestMethods.OPTIONS);
const Use = (...middlewares) => (...args) => {
    if (args.length === 3)
        Reflect.defineMetadata(enums_1.MetaDataKeys.ROUTE_MIDDLEWARES, [
            ...(Reflect.getMetadata(enums_1.MetaDataKeys.ROUTE_MIDDLEWARES, args[0], args[1]) || []),
            ...middlewares,
        ], args[0], args[1]);
    else
        Reflect.defineMetadata(enums_1.MetaDataKeys.MIDDLEWARES, [
            ...(Reflect.getMetadata(enums_1.MetaDataKeys.MIDDLEWARES, args[0]) || []),
            ...middlewares,
        ], args[0]);
};
exports.Use = Use;
