import { Router, RouterOptions } from "express";
import { EpicHelpers, RequestHandler } from "../lib/helpers";
import { MetaDataKeys, RequestMethods } from "./enums";
import { EpicGlobals } from "../lib/globals";

/**
 *
 * Controller Decorator
 * This decorator allows you to create Controller Classes for your Application.
 *
 */

export interface ControllerInterface {
  prefix: string;
  router: Router;
  constructor: new () => any;
  options: ControllerOptions;
}

export interface ControllerOptions {
  childs?: (new () => any)[];
  isChild?: true;
  isOpen?: true;
  routerOptions?: RouterOptions;
}

export const Controller = (prefix = "/", options?: ControllerOptions) => (
  constructor: new () => any
) => {
  // Get Parent Router from Metadata
  const ChildControllers: ControllerInterface[] =
    options?.childs instanceof Array
      ? [
          ...options.childs
            .map<ControllerInterface>((child) =>
              Reflect.getMetadata(MetaDataKeys.CONTROLLER, child)
            )
            .filter((v) => v),
        ]
      : [];

  // Create Target Router
  const TargetRouter = Router(options?.routerOptions);

  // Save Controller to Metadata
  Reflect.defineMetadata(
    MetaDataKeys.CONTROLLER,
    {
      prefix,
      router: TargetRouter,
      constructor,
      options,
    },
    constructor
  );

  // Add Middlewares
  const Middlewares =
    Reflect.getMetadata(MetaDataKeys.MIDDLEWARES, constructor) || [];
  if (Middlewares.length) TargetRouter.use(...Middlewares);

  // Push Child Controllers
  ChildControllers.forEach((controller) =>
    TargetRouter.use(controller.prefix, controller.router)
  );

  // Add Sub Router To Main Or Parent Router Instance
  if (!options?.isChild) EpicGlobals.getRouter().use(prefix, TargetRouter);

  // Get Routes
  const Routes: RouteHandler[] =
    Reflect.getMetadata(MetaDataKeys.ROUTES, constructor.prototype) || [];

  // Loop over Each Route
  Routes.forEach((Route) => {
    // Push Route Handler Details To Globals
    EpicGlobals.pushRouteDetails({
      name: Route.name,
      description: Route.description,
    });

    // Add Route to Application
    TargetRouter[Route.method](
      Route.path,
      // Add Request Name
      (req, _, next) => {
        // @ts-ignore
        req.name = Route.name;
        next();
      },
      // Add Route Middlewares
      ...(Reflect.getMetadata(
        MetaDataKeys.ROUTE_MIDDLEWARES,
        constructor.prototype,
        Route.name
      ) || []),
      // Check Permissions
      EpicHelpers.requestHandler((req, res, next) => {
        // @ts-ignore
        if (!options?.isOpen && req.permissions instanceof Array) {
          // @ts-ignore
          if (!req.permissions.includes(req.name)) {
            res.status(401);
            return next(new Error(`You are not permitted!`));
          }
        }

        next();
      }),
      EpicHelpers.requestHandler(Route.handler)
    );
  });
};

export interface ChildControllerOptions {
  childs?: (new () => any)[];
  routerOptions?: RouterOptions;
}

export const ChildController = (prefix = "/", options?: ControllerOptions) =>
  Controller(prefix, { ...options, isChild: true });

/**
 *
 * Routes Decorator
 * This decorator allows you to create routes under Controller.
 *
 */

export interface RouteHandler {
  method: RequestMethods;
  path: string;
  name: string;
  description: string;
  handler: RequestHandler;
}

export interface RouteHandlerDescriptor extends PropertyDescriptor {
  value?: RequestHandler;
}

// Create Route Binder Function
const RouteBinder = (method: RequestMethods) => (
  path = "/",
  description = "No description has been provided!"
) => (target: any, key: string, desc: RouteHandlerDescriptor) => {
  if (typeof desc.value === "function")
    Reflect.defineMetadata(
      MetaDataKeys.ROUTES,
      [
        ...(Reflect.getMetadata(MetaDataKeys.ROUTES, target) || []),
        {
          path,
          method,
          name: key,
          description,
          handler: desc.value,
        },
      ],
      target
    );
};

// Available Request Method Decorators
export const Get = RouteBinder(RequestMethods.GET);
export const Post = RouteBinder(RequestMethods.POST);
export const Patch = RouteBinder(RequestMethods.PATCH);
export const Put = RouteBinder(RequestMethods.PUT);
export const Delete = RouteBinder(RequestMethods.DELETE);
export const Options = RouteBinder(RequestMethods.OPTIONS);

/**
 *
 * Middleware Decorator
 * This decorator allows you to add middlewares on the route.
 *
 */

export interface UseDecorator {
  (constructor: new () => any): void;
  (target: any, key: string, desc: RouteHandlerDescriptor): void;
}

export const Use = (...middlewares: RequestHandler[]): UseDecorator => (
  ...args: any[]
) => {
  if (args.length === 3)
    Reflect.defineMetadata(
      MetaDataKeys.ROUTE_MIDDLEWARES,
      [
        ...(Reflect.getMetadata(
          MetaDataKeys.ROUTE_MIDDLEWARES,
          args[0],
          args[1]
        ) || []),
        ...middlewares,
      ],
      args[0],
      args[1]
    );
  else
    Reflect.defineMetadata(
      MetaDataKeys.MIDDLEWARES,
      [
        ...(Reflect.getMetadata(MetaDataKeys.MIDDLEWARES, args[0]) || []),
        ...middlewares,
      ],
      args[0]
    );
};
