import express, { Request, Response } from "express";
import { HTTPBase } from "./http-base";
import { EpicServerException } from "./exceptions";
import { CreateResponse, RequestErrorHandler } from "./helpers";

/**
 *
 * Create Epic Application
 * This class will create an instance of Application framework. Default framework is express.
 *
 */

export interface EpicApplicationInterface {
  _beforeInit?: (Server: HTTPBase) => Promise<any> | any;
  _onInit?: (Server: HTTPBase) => Promise<any> | any;
  _onRouteError?: RequestErrorHandler;
  _onError?: (
    error: EpicServerException,
    _Server: HTTPBase
  ) => Promise<any> | any;
}

export class EpicApplication {
  constructor(
    options?: EpicApplicationInterface,
    public Framework = express()
  ) {
    this._beforeInit = options?._beforeInit || this._beforeInit;
    this._onInit = options?._onInit || this._onInit;
    this._onRouteError = options?._onRouteError || this._onRouteError;
    this._onError = options?._onError || this._onError;
  }

  /**
   * Before the server initializes
   * This method let you configure the middleware required by your application to works.
   *
   */
  public _beforeInit = (_Server: HTTPBase): Promise<any> | any => {};

  /**
   * When the server starts listening
   * Do something when the server starts listening...
   *
   */
  public _onInit = (Server: HTTPBase): Promise<any> | any => {
    // Log Server Initialization
    console.log(`Server is listening at Port: ${Server.Port}.`);
  };

  /**
   * When the server starts listening
   * Do something when the server starts listening...
   *
   */
  public _onRouteError: RequestErrorHandler = (
    err: any,
    _req: Request,
    res: Response
  ) => {
    // Set Error Status
    res.status(500);

    // Return Response Object
    return new CreateResponse(err);
  };

  /**
   *
   * Error occured while starting the server
   * Do something if there was an error while starting the server.
   *
   */
  public _onError = (
    error: EpicServerException,
    _Server: HTTPBase
  ): Promise<any> | any => {
    // Log Error
    console.error(error);

    // Exit Application
    process.exit(1);
  };
}
