import { Request, Response, NextFunction } from "express";

export interface ValidationMessage {
  param?: string;
  location?: string;
  message: string;
  value?: any;
}

export interface DefaultResponse<B extends boolean, T> {
  status: B;
  messages: Array<ValidationMessage>;
  data: T;
  code: number;
}

export class CreateResponse {
  protected Status = true;
  protected Code?: number;

  constructor(protected Message: string, protected Data?: any) {}

  public isFalse() {
    this.Status = false;
    return this;
  }

  public setCode(code: number) {
    this.Code = code;
    return this;
  }

  public render() {
    return [this.Status, this.Message, this.Data, this.Code] as const;
  }
}

export type RequestHandlerReturnType = CreateResponse | string | void;

export type RequestHandler = (
  req: Request<any, any, any, any, any>,
  res: Response<any, any>,
  next: NextFunction
) => Promise<RequestHandlerReturnType> | RequestHandlerReturnType;

export type RequestErrorHandler = (
  err: any,
  req: Request<any, any, any, any, any>,
  res: Response<any, any>
) => Promise<RequestHandlerReturnType> | RequestHandlerReturnType;

export class EpicHelpers {
  static requestHandler(Handler: RequestHandler) {
    // Return Middleware
    return async (req: Request, res: Response, next: NextFunction) => {
      try {
        // Process Controller Operations
        const Results = await Handler(req, res, next);

        if (Results)
          // Return Response
          res.json(
            await this.response(
              ...(Results instanceof CreateResponse
                ? Results
                : new CreateResponse(Results)
              ).render()
            )
          );
      } catch (error) {
        next(error);
      }
    };
  }

  static requestErrorHandler(Handler: RequestErrorHandler) {
    // Return Middleware
    return async (
      err: any,
      req: Request,
      res: Response,
      _next: NextFunction
    ) => {
      // Process Controller Operations
      const TempResults = await Handler(err, req, res);

      // Resolve Results
      const Results = (TempResults instanceof CreateResponse
        ? TempResults
        : new CreateResponse(
            TempResults || "An unexpected error has been encountered!"
          )
      ).render();

      // Return Response
      res.json(
        await this.response(
          false,
          Results[1],
          Results[2] || (process.env.NODE_ENV === "development" ? err : {}),
          Results[3]
        )
      );
    };
  }

  static async response<B extends boolean, T>(
    status: B,
    messages: string | ValidationMessage | ValidationMessage[],
    data?: T,
    code?: number
  ): Promise<DefaultResponse<B, T>> {
    // Resolve Messages
    if (typeof messages !== "object" && typeof messages !== "string")
      messages = status
        ? "The request was successful!"
        : "The request was failed!";

    // Format the Message
    if (typeof messages === "string") messages = { message: messages };
    else if (messages instanceof Error)
      messages = { message: messages.message };

    // Return Data
    return {
      status,
      messages: messages instanceof Array ? messages : [messages],
      data: data || ({} as T),
      code: code || (status ? 0 : 1),
    };
  }
}
