import http from "http";
import https from "https";
import createError from "http-errors";
import RequestID from "express-request-id";
import { Debugger } from "debug";
import { EpicApplication } from "./application";
import { EpicGlobals } from "./globals";
import { EpicServerException } from "./exceptions";
import { EpicHelpers } from "./helpers";

export class HTTPBase {
  public Port!: number;

  constructor(
    public Application: EpicApplication,
    public Server: http.Server | https.Server,
    public ServerDebug: Debugger
  ) {}

  public async listen(port: string | number) {
    // Normalize Port
    this.Port = this.normalizePort(port);

    // Add Request ID
    this.Application.Framework.use(RequestID());

    // Before Start Listening to the Server
    await this.Application._beforeInit(this);

    // Add Routes
    this.Application.Framework.use(EpicGlobals.getRouter());

    // Add Routes Error Handler
    this.Application.Framework.use((_req, _res, next) =>
      next(createError(404))
    ).use(EpicHelpers.requestErrorHandler(this.Application._onRouteError));

    // Listen on provided port, on all network interfaces.
    this.Server.listen(this.Port);

    // Event listener for HTTP/HTTPS Server.
    this.Server.on("listening", async () => {
      const Address = this.Server.address();
      const Bind =
        typeof Address === "string"
          ? "Pipe " + Address
          : "Port " + Address?.port;

      this.ServerDebug("Listening on " + Bind);

      await this.Application._onInit(this);
    });

    // Event listener for HTTP server "error" event.
    this.Server.on("error", async (error: any) => {
      try {
        if (error.syscall !== "listen") throw error;

        // Get Port Binding
        const Bind =
          typeof this.Port === "string"
            ? "Pipe " + this.Port
            : "Port " + this.Port;

        // Handle specific listen errors with friendly messages
        switch (error.code) {
          case "EACCES":
            throw new Error(Bind + " requires elevated privileges!");
          case "EADDRINUSE":
            throw new Error(Bind + " is already in use!");
          default:
            throw error;
        }
      } catch (error) {
        // Error Occured While Starting the Server
        await this.Application._onError(
          new EpicServerException((error as any).message || error),
          this
        );
      }
    });
  }

  /**
   *
   * Utility Functions
   * These utility functions are used to safely start the server.
   *
   */

  // Normalize Server Port
  public normalizePort(port: number | string) {
    // Resolve Port
    if (typeof port == "string") port = parseInt(port, 10);

    // Validate Port
    if (isNaN(port)) return port;
    else if (port >= 0) return port;
    else
      throw new EpicServerException(
        `Invalid Port '${port}' has been provided!`
      );
  }
}
